{
-- alex Lexer.x
-- ghci -Wall Lexer.hs

module Lexer  (Token(..), Result(..), Lexer(..), get, srcLoc, execute, panic)
where
import Char
import AbstractSyntax (Expr, Loc(..))

}

$white  = [\ \t\n\v\f\r]
$digit  = 0-9                                          -- digits
$upper  = [A-Z\196\214\220]                            -- upper-case (A)
$lower  = [a-z\228\246\252\223]                        -- lower-casee ()
$letter = [$lower$upper]                               -- letters
-- $symbol = [$letter!\$\%&\'\*\+\,\-\/\<=>\?@\^_`\|]  -- symbols
$symbol = [$letter$digit!\$\%\'\-\?@\^_`] -- symbols
-- &\*\+\/\<=>

tokens :-

-- white characters
$white+                         ;
-- single-line comment (a la C) [ NB. no longer: must be terminated with a newline, not the EOF! ]
\/\/[^\n]*                      ;
-- nested comments (a la C, except that C doesn't support nested comment)
\/\*                            { \ s l -> nestedComment         }
-- reserved words
abstract                        { \ s l -> return ABSTRACT       }
and                             { \ s l -> return AND            }
array                           { \ s l -> return ARRAY          }
case                            { \ s l -> return CASE           }
catch                           { \ s l -> return CATCH          }
class                           { \ s l -> return CLASS          }
con                             { \ s l -> return CON            }
--cont                            { \ s l -> return CONT           }
continue                        { \ s l -> return CONTINUE       }
data                            { \ s l -> return DATA           }
do                              { \ s l -> return DO             }
else                            { \ s l -> return ELSE           }
end                             { \ s l -> return END            }
exception                       { \ s l -> return EXCEPTION      }
-- false                           { \ s l -> return FALSE          }
-- fi                              { \ s l -> return FI             }
forall                          { \ s l -> return FORALL         }
fun                             { \ s l -> return FUN            }
function                        { \ s l -> return FUNCTION       }
functor                         { \ s l -> return FUNCTOR        }
if                              { \ s l -> return IF             }
in                              { \ s l -> return IN             }
include                         { \ s l -> return INCLUDE        }
interface                       { \ s l -> return INTERFACE      }
inherit                         { \ s l -> return INHERIT        }
let                             { \ s l -> return LET            }
letcc                           { \ s l -> return LETCC          }
local                           { \ s l -> return LOCAL          }
method                          { \ s l -> return METHOD         }
module                          { \ s l -> return MODULE         }
-- name                            { \ s l -> return NAME           }
new                             { \ s l -> return NEW            }
-- od                              { \ s l -> return OD             }
object                          { \ s l -> return OBJECT         }
of                              { \ s l -> return OF             }
open                            { \ s l -> return OPEN           }
rec                             { \ s l -> return REC            }
ref                             { \ s l -> return REF            }
-- show                            { \ s l -> return SHOW           }
sig                             { \ s l -> return SIG            }
struct                          { \ s l -> return STRUCT         }
then                            { \ s l -> return THEN           }
throw                           { \ s l -> return THROW          }
-- to                              { \ s l -> return TO             }
-- true                            { \ s l -> return TRUE           }
try                             { \ s l -> return TRY            }
type                            { \ s l -> return TYPE           }
val                             { \ s l -> return VAL            }
-- value                           { \ s l -> return VALUE          }
while                           { \ s l -> return WHILE          }
-- write                           { \ s l -> return WRITE          }
-- separators
\(                              { \ s l -> return LPAREN         }
\)                              { \ s l -> return RPAREN         }
\,                              { \ s l -> return COMMA          }
\.                              { \ s l -> return PERIOD         }
:                               { \ s l -> return COLON          }
\;                              { \ s l -> return SEMICOLON      }
\[                              { \ s l -> return LBRACKET       }
\\                              { \ s l -> return BACKSLASH      }
\]                              { \ s l -> return RBRACKET       }
\{                              { \ s l -> return LBRACE         }
\}                              { \ s l -> return RBRACE         }
\~                              { \ s l -> return TILDE          }
-- reserved symbols
!                               { \ s l -> return EXCLAMATION    }
\%                              { \ s l -> return PERCENT        }
&                               { \ s l -> return AMPERSAND      }
\<                              { \ s l -> return LANGLE         }
=                               { \ s l -> return EQUAL          }
>                               { \ s l -> return RANGLE         }
\?                              { \ s l -> return (VARID s l)    }
\*                              { \ s l -> return ASTERISK       }
\+                              { \ s l -> return PLUS           }
\-                              { \ s l -> return HYPHEN         }
\/                              { \ s l -> return SLASH          }
\^                              { \ s l -> return CIRCUMFLEX     }
_                               { \ s l -> return UNDERSCORE     }
\|                              { \ s l -> return BAR            }
-- multiple character symbols
&&                              { \ s l -> return CONJUNCTION            }
\|\|                            { \ s l -> return DISJUNCTION             }
==                              { \ s l -> return ISEQUAL        }
:=                              { \ s l -> return BECOMES        }
\/=                             { \ s l -> return UNEQUAL        }
\->                             { \ s l -> return RARROW         }
=>                              { \ s l -> return RDOUBLEARROW   }
>\#                             { \ s l -> return GREATERUNEQUAL }
\#\<                            { \ s l -> return LESSUNEQUAL    }
>=                              { \ s l -> return GREATEREQUAL   }
=\<                             { \ s l -> return LESSEQUAL      }
-- identifiers
-- [$lower\$\%\'@\^_`][$symbol]*   { \ s l -> return (VARID s   l)  }
[$lower][$symbol]*              { \ s l -> return (if last s == '`' then PREVARID s l else VARID  s l) }
`[$lower][$symbol]*             { \ s l -> return (if last s == '`' then DISTID   s l else POSTID s l) }
[$upper][$symbol]*              { \ s l -> return (if last s == '`' then PRECONID s l else CONID  s l) }
`[$upper][$symbol]*             { \ s l -> return (if last s == '`' then DISTID   s l else POSTID s l) }
-- natural number/numeral
$digit+                         { \ s l -> return (NAT (read s) l) }
-- string literal (essentially ANSI-C strings)
\"				{ \ s l -> cString           l   }

{
-- The token type:
data Token =
     ABSTRACT
  |  AND
  |  ARRAY
  |  CASE
  |  CATCH
  |  CLASS
  |  CONTINUE
  |  CON
--  |  CONT
  |  DATA
  |  DO
  |  ELSE
  |  END
  |  EXCEPTION
--  |  FALSE
--  |  FI
--  |  FOR
  |  FORALL
  |  FUN
  |  FUNCTION
  |  FUNCTOR
  |  IF
  |  IN
  |  INCLUDE
  |  INTERFACE
  |  INHERIT
  |  LET
  |  LETCC
  |  LOCAL
  |  METHOD
  |  MODULE
--  |  NAME
  |  NEW
  |  OBJECT
--  |  OD
  |  OF
  |  OPEN
  |  REC
  |  REF
--  |  SHOW
  |  SIG
  |  STRUCT
  |  THEN
  |  THROW
--  |  TO
--  |  TRUE
  |  TRY
  |  TYPE
  |  VAL
--  |  VALUE
  |  WHILE
--  |  WRITE

  |  EXCLAMATION
  |  PERCENT
  |  AMPERSAND
  |  LPAREN
  |  RPAREN
  |  ASTERISK
  |  PLUS
  |  COMMA
  |  HYPHEN
  |  PERIOD
  |  SLASH
  |  COLON
  |  SEMICOLON
  |  LANGLE
  |  EQUAL
  |  RANGLE
  |  LBRACKET
  |  BACKSLASH
  |  RBRACKET
  |  CIRCUMFLEX
  |  UNDERSCORE
  |  LBRACE
  |  BAR
  |  RBRACE
  |  TILDE

  |  DISJUNCTION
  |  CONJUNCTION
  |  ISEQUAL
  |  BECOMES
  |  UNEQUAL
  |  RARROW
  |  RDOUBLEARROW  
  |  GREATERUNEQUAL
  |  LESSUNEQUAL
  |  GREATEREQUAL
  |  LESSEQUAL

  |  VARID         String   Loc
  |  CONID         String   Loc
  |  PREVARID      String   Loc
  |  PRECONID      String   Loc
  |  DISTID        String   Loc
  |  POSTID        String   Loc

  |  NAT           Integer  Loc
  |  STRING        String   Loc

  |  EOF
     deriving (Show)

-- lexing actions

type Action = String -> Lexer Token

data Result a = Fail String | Succeed a
                deriving (Show)

instance Monad Result where
  Fail s    >>=  _k  =  Fail s
  Succeed a >>=  k   =  k a
  return             =  Succeed
  fail               =  Fail

-- hand-written lexer for nested comments

nestedComment :: Lexer Token
nestedComment
  =  do  inp <- getInput
         case nested 1 inp of
           Nothing    ->  lexicalError "missing `*/'"
           Just inp'  ->  setInput inp' >> get

nested :: Int -> AlexInput -> Maybe AlexInput
nested 0 inp  =  Just inp
nested n inp  =  case alexGetChar inp of
                 Nothing           ->  Nothing
                 Just ('*', inp')  ->  case alexGetChar inp' of
                                       Nothing            ->  Nothing
                                       Just ('/', inp'')  ->  nested (n - 1) inp''
                                       Just (_,   inp'')  ->  nested n       inp''
                 Just ('/', inp')  ->  case alexGetChar inp' of
                                       Nothing            ->  Nothing
                                       Just ('*', inp'')  ->  nested (n + 1) inp''
                                       Just (_,   inp'')  ->  nested n       inp''
                 Just (_,   inp')  ->  nested n inp'

-- hand-written lexer for strings
--  fehlende `"' sind schwer zu erkennen; typischerweise wird das nchste Anfhrungszeichen
--  als Ausfhrungszeichen misinterpretiert; das folgende Stringliteral wird dann als
--  Programm aufgefasst und verursacht typischerweise einen lexikalischen oder einen
--  Syntaxfehler.

cString :: Loc -> Lexer Token
cString l
  =  do  inp <- getInput
         case rawString inp of
           (Fail err,   inp')  ->  setInput inp' >> lexicalError ("malformed string literal: " ++ err)
           (Succeed x,  inp')  ->  setInput inp' >> return (STRING x l)

rawString, escape, control, stringGap :: AlexInput -> (Result [Char], AlexInput)
rawString inp
  =  case alexGetChar inp of
     Nothing            ->  (Fail "missing `\"'", inp)
     Just ('\"', inp')  ->  (Succeed [], inp') -- end of string
     Just ('\\', inp')  ->  escape inp'
     Just (c,    inp')  ->  c <| rawString inp'

escape inp
  =  case alexGetChar inp of
     Nothing            ->  (Fail "escape sequence expected", inp)
     Just ('f',  inp')  ->  '\f' <| rawString inp' -- Seitenvorschub
     Just ('n',  inp')  ->  '\n' <| rawString inp' -- Zeilenumbruch
     Just ('r',  inp')  ->  '\r' <| rawString inp' -- Carriage Return/Wagenvorlauf
     Just ('t',  inp')  ->  '\t' <| rawString inp' -- horizontaler Tabulator
     Just ('v',  inp')  ->  '\v' <| rawString inp' -- vertikaler Tabulator
     Just ('^',  inp')  ->  control inp'
     Just ('"',  inp')  ->  '"'  <| rawString inp'
     Just ('\\', inp')  ->  '\\' <| rawString inp'
     Just (c,    inp')
       | isDigit c      ->  decimal (digit c) inp'
       | isSpace c      ->  stringGap inp'
       | otherwise      ->  (Fail ("illegal escape sequence: `" ++ [c, '\'']), inp)

control inp
  =  case alexGetChar inp of
     Nothing                   ->  (Fail "control character expected", inp)
     Just (c, inp')
       | '@' <= c && c <= '_'  ->  toEnum (fromEnum c - fromEnum '@') <| rawString inp'  -- "@A-Z[\]^_"
       | otherwise             ->  (Fail ("illegal control character: `" ++ [c, '\'']), inp)

decimal :: Int -> AlexInput -> (Result [Char], AlexInput)
decimal n inp
  =  case alexGetChar inp of
     Nothing        ->  (Fail "missing `\"'", inp)
     Just (c, inp')
       | isDigit c  ->  decimal (10 * n + digit c) inp'
       | n <= 255   ->  toEnum n <| rawString inp
       | otherwise  ->  (Fail ("ASCII code out of range: " ++ show n), inp)

stringGap inp
  =  case alexGetChar inp of
     Nothing            ->  (Fail "unterminated string gap", inp)
     Just ('\\', inp')  ->  rawString inp'
     Just (c,    inp')
       | isSpace c      ->  stringGap inp'
       | otherwise      ->  (Fail ("illegal character in string gap: `" ++ [c, '\'']), inp)
       
digit :: Char -> Int
digit c = fromEnum c - fromEnum '0'    

(<|) :: a -> (Result [a], b) -> (Result [a], b)
_a <| (Fail err,   inp)  =  (Fail err,         inp)
a  <| (Succeed x,  inp)  =  (Succeed (a : x),  inp)

-- lexer monad

type CPS a ans  =  (a -> ans) -> ans

--type Answer = Expr
--type Answer = ()

data Lexer a  =  Lexer { apply :: forall answer . CPS a (  AlexInput            -- input
                                           -> Result answer) }  -- final result of the parser
-- NB. This is the only place where a non-Haskell 98 construct is used.
-- Old NB. The answer type is fixed since the real type (forall ans. CPS a ans) is not H98.

instance Monad Lexer where
  return a  =  Lexer (\ cont -> cont a)
  m >>= k   =  Lexer (\ cont -> apply m (\ a -> apply (k a) cont))
  fail s    =  lift (fail s)

emit t = Lexer (\ cont inp@(Input _ _ l c) -> cont (t (At l c)) inp)

lift :: Result a -> Lexer a
lift m  =  Lexer (\cont inp -> m >>= \ a -> cont a inp)

getInput :: Lexer AlexInput
getInput = Lexer (\ cont inp -> cont inp inp)
setInput :: AlexInput -> Lexer ()
setInput inp = Lexer (\ cont _inp -> cont () inp)

srcLoc :: Lexer Loc
srcLoc = Lexer (\ cont inp@(Input _ _ l c) -> cont (At l c) inp)

execute :: Lexer answer -> String -> Result answer
execute m inp = apply m (\ a _ -> return a) (alexInput inp)

-- Alex interface: low-level interface to the input; used both by
--   alex and the hand-written lexers (for comments and strings)

data AlexInput = Input [String]  -- left context (at most 2 lines)
                       String    -- remaining input
                       !Int      -- line number 1..
                       !Int      -- column number 1..

alexInput :: String -> AlexInput
alexInput x = Input [] x 1 0

alexString :: AlexInput -> String
alexString (Input _ x _ _) = x

alexGetChar :: AlexInput -> Maybe (Char, AlexInput)
alexGetChar (Input _ctx [] _l _c) = Nothing
alexGetChar (Input ctx (a : x) l c)
  =  Just (a, if a == '\n' then Input (take 2 ([] : ctx)) x (l + 1) 0
              else if a == '\t' then Input (a <: ctx) x l ((c `div` 8 + 1) * 8)
              else Input (a <: ctx) x l (c + 1))

(<:)  :: a -> [[a]] -> [[a]]
a <: []        =  [[a]]
a <: (x : xs)  =  (a : x) : xs

alexInputPrevChar :: AlexInput -> Char
alexInputPrevChar (Input [] _ _ _)                 =  '\n' -- beginning of input
alexInputPrevChar (Input ([] : _ctx) _ _ _)        =  '\n' -- beginning of line
alexInputPrevChar (Input ((a : _x) : _ctx) _ _ _)  =  a

-- error reporting

lexicalError :: String -> Lexer a
lexicalError msg = panic ("lexical error (" ++ msg ++ ")") ""

panic :: String -> String -> Lexer a
panic msg diagnosis
  = do (Input ctx x l c) <- getInput
--       let c' = 1 `max` c - 1
       fail (msg ++ " at "
             ++ show (At l c) ++ ":\n"
             ++ context c x ctx
             ++ diagnosis)

context :: Int -> String -> [String] -> String
context col inp ctx  =  unlines (above ctx
                                 ++ [replicate col ' ' ++ "^"]
                                 ++ take 2 (lines (tail' inp') ++ ["<end of input>"]))
  where (rest, inp') = break (=='\n') inp
        above []        =  [rest]
        above (x : xs)  =  reverse ((reverse x ++ rest) : map reverse xs)

tail' :: [a] -> [a]
tail' []       =  []
tail' (_ : x)  =  x

-- the real lexer (used by the parser)

get :: Lexer Token
get =  do inp@(Input _ _ r c) <- getInput
          case alexScan inp 0 of
            AlexEOF              -> return EOF
            AlexError inp'       -> setInput inp' >> lexicalError "illegal character"
            AlexSkip  inp' _l    -> setInput inp' >> get
            AlexToken inp' l act -> setInput inp' >> act (take l (alexString inp)) (At r c)

-- TODO: als Position (|Loc|) eines Tokens die Position des ersten Zeichens nehmen.
-- Zur Zeit: Position des letzten Zeichens.

-- requires: Answer = [Token]
-- test = get >>= \ t -> case t of EOF -> return [EOF]
--                                 _   -> test >>= \ ts -> return (t : ts)

}